@@ -19,36 +19,35 @@ const DEFAULT_INITIAL_MAX_STREAMS_UNI: u64 = 100;
 pub struct QuicListener {
     pub socket: mio::net::UdpSocket,
     pub connection: Option<Pin<Box<quiche::Connection>>>,
-    pub is_server: bool,
+    poll: mio::Poll,
+    pub addr: SocketAddr
 }
 
 impl QuicListener {
     pub fn new(addr: SocketAddr) -> io::Result<QuicListener> {
         let socket = net::UdpSocket::bind(addr).unwrap();
         let socket = mio::net::UdpSocket::from_socket(socket).unwrap();
+        let poll = mio::Poll::new().unwrap();
+        poll.register(
+            &socket,
+            mio::Token(0),
+            mio::Ready::readable(),
+            mio::PollOpt::edge(),
+        )
+        .unwrap();
         Ok(QuicListener {
             socket,
             connection: None,
-            is_server: false,
+            poll,
+            addr
         })
     }
 
-    pub fn recv_from(&self) -> Result<(usize, SocketAddr), io::Error> {
+    pub fn recv(&mut self, stream_out: &mut [u8], stream_id: u64) -> Result<usize, io::Error> {
+        // initialise needed buffers
         let mut buf = [0; 65535];
-        loop {
-            match self.socket.recv_from(&mut buf) {
-                Ok((len, from)) => return Ok((len, from)),
-                Err(_) => {
-                    return Err(io::Error::new(
-                        io::ErrorKind::WouldBlock,
-                        "recv would block",
-                    ))
-                }
-            };
-        }
-    }
-
-    pub fn accept(&mut self) -> Result<(), io::Error> {
+        let mut out = [0; DEFAULT_MAX_DATAGRAM_SIZE];
+        // get quiche config
         let mut config = match self.default_quiche_config() {
             Ok(conf) => conf,
             Err(_) => {
@@ -58,27 +57,24 @@ impl QuicListener {
                 ))
             }
         };
+        // for server config, load crt and key
         config.load_cert_chain_from_pem_file("cert.crt").unwrap();
         config.load_priv_key_from_pem_file("cert.key").unwrap();
 
         config
             .set_application_protos(b"\x0ahq-interop\x05hq-29\x05hq-28\x05hq-27\x08http/0.9")
             .unwrap();
-        let poll = mio::Poll::new().unwrap();
+        // set up event loop
         let mut events = mio::Events::with_capacity(1024);
-        let mut buf = [0; 65535];
-        let mut out = [0; DEFAULT_MAX_DATAGRAM_SIZE];
+
+        // create a seed to generate connection id
         let rng = SystemRandom::new();
-        poll.register(
-            &self.socket,
-            mio::Token(0),
-            mio::Ready::readable(),
-            mio::PollOpt::edge(),
-        )
-        .unwrap();
         let conn_id_seed = ring::hmac::Key::generate(ring::hmac::HMAC_SHA256, &rng).unwrap();
+        let mut some_read = None;
+        let mut some_fin = None;
         loop {
-            poll.poll(&mut events, None).unwrap();
+            self.poll.poll(&mut events, None).unwrap();
+            // read incoming UDP packets from the socket and process them
             'read: loop {
                 if events.is_empty() {
                     break 'read;
@@ -93,109 +89,135 @@ impl QuicListener {
                     }
                 };
                 let packet = &mut buf[..len];
+                // parse the QUIC header from the received packet
                 let header = match quiche::Header::from_slice(packet, quiche::MAX_CONN_ID_LEN) {
                     Ok(v) => v,
                     Err(_) => {
                         continue 'read;
                     }
                 };
+
+                // generate a connection id from seed
                 let conn_id = ring::hmac::sign(&conn_id_seed, &header.dcid);
                 let conn_id = &conn_id.as_ref()[..quiche::MAX_CONN_ID_LEN];
+                if self.connection.is_none() {
+                    // version negotiation if version is not supported by the server
+                    if !quiche::version_is_supported(header.version) {
+                        let len = quiche::negotiate_version(&header.scid, &header.dcid, &mut out)
+                            .unwrap();
 
-                if header.ty != quiche::Type::Initial {
-                    continue 'read;
-                }
+                        let out = &out[..len];
 
-                // version negotiation
-                if !quiche::version_is_supported(header.version) {
-                    let len =
-                        quiche::negotiate_version(&header.scid, &header.dcid, &mut out).unwrap();
-
-                    let out = &out[..len];
-
-                    if let Err(e) = self.socket.send_to(out, &from) {
-                        if e.kind() == std::io::ErrorKind::WouldBlock {
-                            break;
-                        }
-                        panic!("send() failed: {:?}", e);
-                    }
-                    continue 'read;
-                }
-                let mut scid = [0; quiche::MAX_CONN_ID_LEN];
-                scid.copy_from_slice(&conn_id);
-                let scid = quiche::ConnectionId::from_ref(&scid);
-                let token = header.token.as_ref().unwrap();
-
-                // do a stateless retry if the client didn't send a token
-                if token.is_empty() {
-                    let new_token = mint_token(&header, &from);
-
-                    let len = quiche::retry(
-                        &header.scid,
-                        &header.dcid,
-                        &scid,
-                        &new_token,
-                        header.version,
-                        &mut out,
-                    )
-                    .unwrap();
-                    let out = &out[..len];
-                    if let Err(e) = self.socket.send_to(out, &from) {
-                        if e.kind() == std::io::ErrorKind::WouldBlock {
-                            break;
-                        }
-
-                        panic!("send() failed: {:?}", e);
-                    }
-                    continue 'read;
-                }
-                let odcid = validate_token(&from, token);
-
-                if odcid.is_none() {
-                    continue 'read;
-                }
-
-                if scid.len() != header.dcid.len() {
-                    continue 'read;
-                }
-
-                // reuse the source connection id
-                let scid = header.dcid.clone();
-
-                // new connection
-                let mut conn = quiche::accept(&scid, odcid.as_ref(), from, &mut config).unwrap();
-                if conn.is_established() {
-                    self.connection = Some(conn);
-                    return Ok(());
-                } else {
-                    let recv_info = quiche::RecvInfo { from };
-                    match conn.recv(packet, recv_info) {
-                        Ok(v) => v,
-                        Err(_) => {
-                            continue 'read;
-                        }
-                    };
-                    loop {
-                        let (write, _) = match conn.send(&mut out) {
-                            Ok(v) => v,
-                            Err(quiche::Error::Done) => {
-                                break;
-                            }
-                            Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
-                        };
-                        if let Err(e) = self.socket.send_to(&mut out[..write], &from) {
+                        if let Err(e) = self.socket.send_to(out, &from) {
                             if e.kind() == std::io::ErrorKind::WouldBlock {
                                 break;
                             }
                             panic!("send() failed: {:?}", e);
                         }
+                        continue 'read;
+                    }
+                    // reuse the connection is as a source id
+                    let mut scid = [0; quiche::MAX_CONN_ID_LEN];
+                    scid.copy_from_slice(&conn_id);
+                    let scid = quiche::ConnectionId::from_ref(&scid);
+                    // if there is no connection then we need to create a new one
+                    // token has to be present on initial packets
+                    let token = header.token.as_ref().unwrap();
+
+                    // do a stateless retry if the client didn't send a token
+                    if token.is_empty() {
+                        let new_token = mint_token(&header, &from);
+
+                        let len = quiche::retry(
+                            &header.scid,
+                            &header.dcid,
+                            &scid,
+                            &new_token,
+                            header.version,
+                            &mut out,
+                        )
+                        .unwrap();
+                        let out = &out[..len];
+                        if let Err(e) = self.socket.send_to(out, &from) {
+                            if e.kind() == std::io::ErrorKind::WouldBlock {
+                                break;
+                            }
+
+                            panic!("send() failed: {:?}", e);
+                        }
+                        continue 'read;
+                    }
+                    let odcid = validate_token(&from, token);
+
+                    // drop packet
+                    if odcid.is_none() {
+                        continue 'read;
+                    }
+
+                    // drop packet
+                    if scid.len() != header.dcid.len() {
+                        continue 'read;
+                    }
+
+                    // reuse the source connection id
+                    let scid = header.dcid.clone();
+
+                    // new connection
+                    let conn = quiche::accept(&scid, odcid.as_ref(), from, &mut config).unwrap();
+                    self.connection = Some(conn);
+                }
+                // Process potentially coalesced packets
+                let mut conn = self.connection.take().unwrap();
+                let recv_info = quiche::RecvInfo { from };
+                let _ = match conn.recv(packet, recv_info) {
+                    Ok(v) => v,
+                    Err(_) => {
+                        continue 'read;
+                    }
+                };
+                if conn.is_established() {
+                    while let Ok((read, fin)) = conn.stream_recv(stream_id, stream_out) {
+                        some_read = Some(read);
+                        some_fin = Some(fin);
                     }
                 }
+                self.connection = Some(conn);
             }
+            // Generate outgoing QUIC packets for connection
+            if self.connection.is_none() {
+                continue;
+            }
+            let mut conn = self.connection.take().unwrap();
+            loop {
+                let (write, send_info) = match conn.send(&mut out) {
+                    Ok(v) => v,
+                    Err(quiche::Error::Done) => {
+                        break;
+                    }
+                    Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
+                };
+                if let Err(e) = self.socket.send_to(&mut out[..write], &send_info.to) {
+                    if e.kind() == std::io::ErrorKind::WouldBlock {
+                        break;
+                    }
+                    panic!("send() failed: {:?}", e);
+                }
+            }
+            if conn.is_established() && some_fin.is_some() {
+                self.connection = Some(conn);
+                return Ok(some_read.unwrap());
+            }
+            self.connection = Some(conn);
         }
     }
 
-    pub fn connect(&mut self, addr: SocketAddr) -> Result<(), io::Error> {
+    pub fn send(
+        &mut self,
+        addr: SocketAddr,
+        payload: &mut [u8],
+        stream_id: u64,
+        fin: bool,
+    ) -> Result<(), io::Error> {
         let mut config = match self.default_quiche_config() {
             Ok(conf) => conf,
             Err(_) => {
@@ -209,28 +231,35 @@ impl QuicListener {
         config
             .set_application_protos(b"\x0ahq-interop\x05hq-29\x05hq-28\x05hq-27\x08http/0.9")
             .unwrap();
-        let scid = quiche::ConnectionId::from_ref(&[0xba; 16]);
-        let poll = mio::Poll::new().unwrap();
         let mut events = mio::Events::with_capacity(1024);
-        let mut conn = quiche::connect(None, &scid, addr, &mut config).unwrap();
-        poll.register(
-            &self.socket,
-            mio::Token(0),
-            mio::Ready::readable(),
-            mio::PollOpt::edge(),
-        )
-        .unwrap();
         let mut buf = [0; 65535];
         let mut out = [0; DEFAULT_MAX_DATAGRAM_SIZE];
-        let (write, send_info) = conn.send(&mut out).expect("initial send failed");
-        while let Err(e) = self.socket.send_to(&out[..write], &send_info.to) {
-            if e.kind() == std::io::ErrorKind::WouldBlock {
-                continue;
-            }
-            panic!("send() failed: {:?}", e);
-        }
+        let mut scid = [0; quiche::MAX_CONN_ID_LEN];
+        SystemRandom::new().fill(&mut scid[..]).unwrap();
+        let scid = quiche::ConnectionId::from_ref(&scid);
+        // generate connection
+        let mut conn = quiche::connect(None, &scid, addr, &mut config).unwrap();
+        // initiate handshake
         loop {
-            poll.poll(&mut events, conn.timeout()).unwrap();
+            let (write, send_info) = match conn.send(&mut out) {
+                Ok(v) => v,
+                Err(quiche::Error::Done) => {
+                    break;
+                }
+                Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
+            };
+            if let Err(e) = self.socket.send_to(&mut out[..write], &send_info.to) {
+                if e.kind() == std::io::ErrorKind::WouldBlock {
+                    break;
+                }
+                panic!("send() failed: {:?}", e);
+            }
+        }
+        self.connection = Some(conn);
+        loop {
+            let mut conn = self.connection.take().unwrap();
+            self.poll.poll(&mut events, None).unwrap();
+            // Read incoming UDP packets from the socket and process them
             'read: loop {
                 if events.is_empty() {
                     break 'read;
@@ -246,6 +275,7 @@ impl QuicListener {
                     }
                 };
                 let recv_info = quiche::RecvInfo { from };
+                // Process potentially coalesced packets
                 let read = match conn.recv(&mut buf[..len], recv_info) {
                     Ok(v) => v,
                     Err(_) => {
@@ -253,49 +283,29 @@ impl QuicListener {
                     }
                 };
                 let packet = &mut buf[..read];
+                // derive header from packet
                 let header = match quiche::Header::from_slice(packet, quiche::MAX_CONN_ID_LEN) {
                     Ok(v) => v,
                     Err(_) => {
                         continue 'read;
                     }
                 };
+                // If we got a retry packet then respond with token that was sent
                 if header.ty == quiche::Type::Retry {
-                    let mut write = None;
+                    let token = header.token.unwrap();
                     loop {
-                        match conn.send(&mut out) {
-                            Ok((len, _)) => write = Some(len),
-                            Err(quiche::Error::Done) => {
-                                break;
-                            }
-                            Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
-                        };
-                    }
-                    let mut initial_with_retry =
-                        [&out[..write.unwrap()], &header.token.unwrap()[..]].concat();
-                    if let Err(e) = self.socket.send_to(&mut initial_with_retry[..], &from) {
-                        if e.kind() == std::io::ErrorKind::WouldBlock {
-                            break;
-                        }
-                        panic!("send() failed: {:?}", e);
-                    }
-                    continue 'read;
-                } else {
-                    let recv_info = quiche::RecvInfo { from };
-                    match conn.recv(packet, recv_info) {
-                        Ok(v) => v,
-                        Err(_) => {
-                            continue 'read;
-                        }
-                    };
-                    loop {
-                        let (write, _) = match conn.send(&mut out) {
+                        let (write, send_info) = match conn.send(&mut out) {
                             Ok(v) => v,
                             Err(quiche::Error::Done) => {
                                 break;
                             }
                             Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
                         };
-                        if let Err(e) = self.socket.send_to(&mut out[..write], &from) {
+                        let mut initial_with_retry = [&out[..write], &token[..]].concat();
+                        if let Err(e) = self
+                            .socket
+                            .send_to(&mut initial_with_retry[..], &send_info.to)
+                        {
                             if e.kind() == std::io::ErrorKind::WouldBlock {
                                 break;
                             }
@@ -303,25 +313,33 @@ impl QuicListener {
                         }
                     }
                 }
-
-                if conn.is_established() {
-                    self.connection = Some(conn);
-                    return Ok(());
+            }
+            // Generate outgoing QUIC packets
+            if conn.is_established() {
+                conn.stream_send(stream_id, payload, fin).unwrap();
+            }
+            loop {
+                let (write, send_info) = match conn.send(&mut out) {
+                    Ok(v) => v,
+                    Err(quiche::Error::Done) => {
+                        break;
+                    }
+                    Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
+                };
+                if let Err(e) = self.socket.send_to(&mut out[..write], &send_info.to) {
+                    if e.kind() == std::io::ErrorKind::WouldBlock {
+                        break;
+                    }
+                    panic!("send() failed: {:?}", e);
                 }
             }
-        }
-    }
-
-    pub fn stream_recv(&mut self, stream_id: u64, out: &mut [u8]) -> io::Result<(usize, bool)> {
-        let mut conn = self.connection.take().unwrap();
-        while let Ok((read, fin)) = conn.stream_recv(stream_id, out) {
-            if fin {
+            // if handshake is complete then connecting is finished
+            if conn.is_established() {
                 self.connection = Some(conn);
-                return Ok((read, fin));
+                return Ok(());
             }
+            self.connection = Some(conn);
         }
-        self.connection = Some(conn);
-        Ok((0, false))
     }
 
     fn default_quiche_config(&self) -> Result<quiche::Config, io::Error> {
